/*
 * tokens.c: Functions dealing with the token stream in a CTF file.
 * Copyright (c) Warren Toomey, under the GPL3 license.
 * 
 * $Revision: 1.14 $
 */

#include <sys/types.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "tokens.h"

/*
 * Open the named CTF file for reading, checking the header as well. Returns
 * the FILE handle to the open file.
 */
FILE * ctfopen(char *name)
{
  FILE *zin = fopen(name, "r");
  if (zin == NULL) {
    fprintf(stderr, "Can't open %s\n", name);
    exit(1);
  }

  /* Check the ctf header */
  if ((fgetc(zin) != 'c') || (fgetc(zin) != 't') || (fgetc(zin) != 'f') ||
      (fgetc(zin) != '2') || (fgetc(zin) != '.') || (fgetc(zin) != '1')) {
    fprintf(stderr, "Input file %s is not a ctf file\n", name);
    exit(1);
  }
  return (zin);
}


/*
 * Given a FILE and an offset, return the next token from the file. The
 * offset is updated to point at the next token. Any filename associated with
 * a FILENAME token is returned, and the id param is used to return the
 * timestamp. Any id-value associated with the the token is returned, or 0 if
 * there is no value. On any error, 0 is returned. The
 * space for the filename is malloc'd here.
 */
unsigned int get_token(FILE * zin, uint32_t *offset, uint32_t *id, char **name)
{
  unsigned int ch, token;
  uint32_t idvalue;
  char *fname;
  int fcnt = 0;			/* # of chars in fname so far */

  /* Give up if zin or offset don't exist */
  if ((zin == NULL) || (offset == NULL))
    return (0);

  /* Get the token */
  (void)fseek(zin, *offset, SEEK_SET);
  token = fgetc(zin);
  if (token == EOF)
    return (0);
  token &= 255;

  /* Set a default id value */
  if (id)
    *id = 0;

  /*
   * Deal with filenames and the id numbers for certain tokens
   */
  switch (token) {
  case STRINGLIT:
  case CHARCONST:
  case LABEL:
  case IDENTIFIER:
  case INTVAL:
    /* Read in 2 bytes to get the id value */
    idvalue = (fgetc(zin) & 0xff) << 8;
    idvalue += (fgetc(zin) & 0xff);
    if (id)
      *id = idvalue;
    break;

  case FILENAME:
#define FNAMESIZE	1000
    /* Read in 4 bytes to get the timestamp value */
    idvalue = (fgetc(zin) & 0xff) << 24;
    idvalue += (fgetc(zin) & 0xff) << 16;
    idvalue += (fgetc(zin) & 0xff) << 8;
    idvalue += (fgetc(zin) & 0xff);
    if (id)
      *id = idvalue;

    /* Make fname */
    fname = malloc(FNAMESIZE + 1);
    if (fname == NULL)
      return (0);

    while ((fcnt < FNAMESIZE) && ((ch = fgetc(zin)) != EOF) && (ch != '\0'))
      fname[fcnt++] = ch;

    /* Error if some chars left */
    if (fcnt >= FNAMESIZE)
      return (0);

    /* Null-terminate it */
    fname[fcnt] = '\0';
    if (name)
      *name = fname;
    else
      free(fname);
  }

  *offset = ftell(zin);
  return (token);
}

/* Nearly the same as above, but skips over LINE tokens */
unsigned int get_nonltoken(FILE * zin, uint32_t *offset, uint32_t *id,
							uint32_t *line)
{
  unsigned int ch;
  while ((ch = get_token(zin, offset, id, NULL)) == LINE)
    (*line)++;
  return (ch);
}

char *tokstring[] = {
  "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ",				/* 0 */
  "ERR ", "ERR ", "\n", "ERR ", "ERR ", ">>= ", "ERR ", "ERR ",					/* 8 */
  "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ",				/* 16 */
  "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ",				/* 24 */
  "/= ", "! ", "\"string\"", "->", "++ ", "% ", "& ", "'c'",					/* 32 */
  "( ", ") ", "* ", "+ ", ", ", "- ", ". ", "/ ",						/* 40 */
  "-- ", "&& ", "|| ", "++= ", "%= ", "-= ", "&= ", "*= ",					/* 48 */
  "|= ", "NUM", ": ", "; ", "< ", "= ", "> ", "? ",						/* 56 */
  "!= ", "<= ", "case ", "char ", "const ", "continue ", "default ", "do ",			/* 64 */
  "...", "double ", "else ", "enum ", "extern ", "float ", "for ", "goto ",			/* 72 */
  "if ", "int ", "long ", "register ", "return", "short ", "signed ", "sizeof ",		/* 80 */
  "static ", "struct ", "switch ", "[ ", "\\", "] ", "^ ", "id",				/* 88 */
  "label: ", "typedef ", "union ", "unsigned ", "void ", "volatile ", "while ", "#define ",	/* 96 */
  "#elif ", "#else ", "#endif ", "#error ", "#ifdef ", "#if ", "#ifndef ", "#include ",		/* 104 */
  "#line ", "#pragma ", "#undef ", "#warning ", "~= ", "== ", "break ", ">= ",			/* 112 */
  "<< ", ">> ", "<<= ", "{ ", "| ", "} ", "~ ", "abstract ",					/* 120 */
  "boolean ", "byte ", "extends ", "final ","finally ","implements ", "import ", "instanceof ", /* 128 */
  "interface ", "native ", "new ", "null ", "package ", "private ", "protected ", "public ",	/* 136 */
  "strictfp ", "super ", "synchronized ", "this ", "throw ", "throws ", "transient ", "try",	/* 144 */
  ">>> ", ">>>= ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ",				/* 152 */
  "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ",				/* 160 */
  "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ",				/* 168 */
  "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ",				/* 176 */
  "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ",				/* 184 */
  "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ",				/* 192 */
  "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ",				/* 200 */
  "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ",				/* 208 */
  "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ",				/* 216 */
  "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ",				/* 224 */
  "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ",				/* 232 */
  "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ",				/* 240 */
  "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR "				/* 248 */
};

char *tok2str(unsigned int ch)
{
  return(tokstring[ch]);
}

void print_token(unsigned int ch, uint32_t linenum, uint32_t id, char *filename)
{
  time_t time;
  switch (ch) {
    case FILENAME:
    time= id;
    printf("\n\n%s:\t%s\n%5d:   ", filename, ctime(&time), linenum);
    break;
  case LINE:
    printf("\n%5d:   ", linenum);
    break;
  case IDENTIFIER:
  case INTVAL:
    printf("%s%d ", tokstring[ch], id);
    break;
  case STRINGLIT:
    printf("\"str%d\" ", id);
    break;
  case LABEL:
    printf("L%d: ", id);
    break;
  case CHARCONST:
    printf("'c%d' ", id);
    break;
  default:
    fputs(tokstring[ch], stdout);
  }
}
